#!/usr/local/bin/perl -w

# SES Server Cron Job.
#
#  remove exports to Linux clients when the client stops refreshing them
#
#  reads export files and things included by them and associated expire files.
#  writes export files.
#

findincs("/etc/exports");

exit(0);

sub findincs {
  my ( $filename ) = @_;
  my $junk; my $nextfile; my $i; my @e;
  my $found = 0;

  print("\nreading $filename\n\n");

  open(IN,$filename) or die "couldn't open $filename: $!";
  @e = <IN>;
  close(IN);

  foreach $i (@e) {
    if ( $i =~ /^.INCLUDE/ ) {
      $found = 1;
      ($junk, $nextfile, $junk) = split(" ",$i);
      findincs($nextfile);
    }
  }
  if ( $found == 0 ) { process($filename) }
}

sub process {
  my ($exportsfile) = @_;
  my $expirefile; my @exports; my @expire;
  my $i; my $dir; my $ip; my $date; my %e; my $now;
  my @bits; my $rest;

  print( "processing $exportsfile\n");

# we are either working directly in etc or with something like
# /cc/ccstaff1/.ses.exports.inc /cc/ccstaff1/.ses.expire
  if ( $exportsfile eq "/etc/exports" ) { $expirefile = "/etc/expire" }
  else { $expirefile = $exportsfile; $expirefile =~ s/exports.inc/expire/ }

# read them both into ram.
  open(IN,$exportsfile) or die "couldn't open $exportsfile: $!";
  @exports = <IN>;
  close(IN);
  open(IN,$expirefile) or die "couldn't open $expirefile: $!";
  @expire = <IN>;
  close(IN);

# store the expire stuff in a hash for easy access
  foreach $i (@expire) {
    ( $dir, $ip, $date ) = split(" ", $i);
# ok, here we have an interesting thing.
# on clnfs0, expire contains /cc/st1/... while exports contains /student1/...
# I'm sure there is a good way to get around this, but for now:
    $dir =~ s/\/cc\/st1\//\/student1\//;
    $dir =~ s/\/cc\/st2\//\/student2\//;
    $dir =~ s/\/cc\/staff1\//\/staff1\//;
    $dir =~ s/\/cc\/staff2\//\/staff2\//;
    $dir =~ s/\/cc\/ccstaff1\//\/ccstaff1\//;

    $e{"$dir:$ip"} = $date;
  }

# what's the time mister wolf
  $now = time();

# read thru exports looking for things to remove
  foreach $i (@exports) {
    next if ( $i =~ /^#/ );
    next if ( $i =~ /^$/ );
    if ( $i =~ /^ / ) { 
      next if ( ! defined($dir) );
      @bits = split(" ",$i);
    }
    else {
      ( $dir, $rest ) = split(" ",$i,2);
      @bits = split(" ",$rest);
    }
    foreach $ip (@bits) {
      next unless ($ip =~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ );
      if ( (defined($e{"$dir:$ip"})) && ( $now - $e{"$dir:$ip"} ) > ( 5 * 60 ) ) {
	remove($exportsfile,$dir,$ip);
      }
    }
  }
}  

sub remove {
  my ($oldfile,$dir,$ip) = @_;
  my $newfile; my $lockfile; my $lin; my @bits; my @thisline; my $i; my $tl;
  my $pnl;

  print("removing $dir $ip from $oldfile\n");

  if ( $oldfile eq "/etc/exports") {
    $newfile = $oldfile . ".new";
    $lockfile = $oldfile . ".lock";
  }
  else {
    $newfile = $oldfile;
    $lockfile = $oldfile;
    $newfile =~ s/.inc/.new/;
    $lockfile =~ s/.inc/.lock/;
  }

  getlock($lockfile);

  if ( !open(IN,"$oldfile") ) { 
    $i = $!; 
    droplock($lockfile); 
    die "can't open $oldfile: $i";
  }
  if ( !open(OUT,">$newfile") ) { 
    $i = $!; 
    close(IN);
    droplock($lockfile); 
    die "can't open $newfile: $i";
  }

  $pnl = 0;
  while ( $lin = <IN> ) {
# compare with dir followed by space so ...1 doesn't match ...11
    if ( index($lin,$dir . " ") == 0 ) {
      @bits = split( " ", $lin);
      undef(@thisline);
      foreach $i (@bits) {
	if ($i ne $ip ) { push(@thisline,$i) }
      }
      while( $lin = <IN> ) {
        if ( $lin =~ /^ / ) {
	  @bits = split( " ", $lin);
	  foreach $i (@bits) {
	    if ($i ne $ip ) { push(@thisline,$i) }
	  }
	}
        else{
	  last;
	}
      }
      $tl = join(" ",@thisline);
      # if at least one IP address remains write it out 
      if ( $tl =~ /[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+/ ) {
        while ( length($tl) > 75 ) {
	  $i = index($tl, " ", 60);
          print(OUT substr($tl,0,$i));
          print(OUT "\n");
	  $tl = substr($tl,$i);
        }
        print(OUT "$tl\n");
        $pnl = 0;
      }
    }
    if ( $lin =~ /^\n/ ) {
      if ( ! $pnl ) {
        print(OUT $lin);
        $pnl = 1;
      }
    }
    else {
      print(OUT $lin);
      $pnl = 0 ;
    }
  }
  close(IN);
  close(OUT);
  unlink($oldfile) or warn "problem unlinking $oldfile: $!";
  link($newfile, $oldfile) or warn "problem linking $oldfile: $!";
  unlink($newfile) or warn "problem unlinking $newfile: $!";

  droplock($lockfile);
}

sub getlock {
  my ($lockfile)=@_;
  my $pidfile; my $i; my $count;

  $pidfile = $lockfile . "." . $$;

  while(1){

    open(PF,">$pidfile") or die "locker can't open pidfile: $!";
    print(PF "$$\n");
    close(PF);
    $count = 10;
    while ( ( ! ( $i = link($pidfile, $lockfile) ) ) && 
            ( $! =~ /File exists/ ) && 
            $count-- ) {
      sleep 1;
    }
    if ( $i != 1 ) {die "locker can't link $pidfile, $lockfile: $!" }
    unlink($pidfile);
    open(IN, "$lockfile") or die "locker can't open lockfile: $!";
    $i = <IN>;
    close(IN);
    last if ($i == $$);
  }
}

sub droplock {
  my ($f)=@_;
  my $pid;

  open(IN,$f) or warn "big stuffup - droplock called but no lockfile";
  $pid = <IN>;
  close(IN);

  if ( $pid != $$ ) { warn "big stuffup - wrong pid in lockfile" }

  unlink($f);
}
